@@ -2,7 +2,7 @@
  *
  * COMMON.H - NRPE Common Include File
  * Copyright (c) 1999-2007 Ethan Galstad (nagios@nagios.org)
- * Last Modified: 11-11-2011
+ * Last Modified: 09-06-2013
  *
  * License:
  *
@@ -23,8 +23,8 @@
 
 #include "config.h"
 
-#define PROGRAM_VERSION "2.13"
-#define MODIFICATION_DATE "11-11-2011"
+#define PROGRAM_VERSION "2.15"
+#define MODIFICATION_DATE "09-06-2013"
 
 #define OK		0
 #define ERROR		-1
@@ -71,7 +71,7 @@ typedef struct packet_struct{
         }packet;
 
 /**************** OPERATING SYSTEM SPECIFIC DEFINITIONS **********/
-#ifdef __sun
+#if defined(__sun) || defined(__hpux)
 
 #  ifndef LOG_AUTHPRIV
 #    define LOG_AUTHPRIV LOG_AUTH
@@ -81,4 +81,12 @@ typedef struct packet_struct{
 #    define LOG_FTP LOG_DAEMON
 #  endif
 
+#elif _AIX
+
+#  include <sys/select.h>
+
+#  ifndef LOG_FTP
+#    define LOG_FTP LOG_DAEMON
+#  endif
+
 #endif
